﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2012 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using System.Linq;
using HeuristicLab.Algorithms.DataAnalysis;
using HeuristicLab.Common;
using HeuristicLab.Problems.DataAnalysis;

namespace HeuristicLab.Algorithms.NCA {
  internal class Auxiliary {

    public static INCAModel Train(IClassificationProblemData data, int k, int dimensions, INCAInitializer initializer) {
      var instances = data.TrainingIndices.Count();
      var attributes = data.AllowedInputVariables.Count();

      double[] matrix = initializer.Initialize(data, dimensions);

      alglib.mincgstate state;
      alglib.mincgreport rep;

      alglib.mincgcreate(matrix, out state);
      alglib.mincgsetcond(state, 0, 0, 0, 20);
      alglib.mincgoptimize(state, Gradient, null, new OptimizationInfo(data, dimensions));
      alglib.mincgresults(state, out matrix, out rep);

      var transformationMatrix = new double[attributes, dimensions];
      var counter = 0;
      for (var i = 0; i < attributes; i++)
        for (var j = 0; j < dimensions; j++)
          transformationMatrix[i, j] = matrix[counter++];

      var transformedTrainingset = new double[instances, dimensions];
      var rowCount = 0;
      foreach (var r in data.TrainingIndices) {
        var i = 0;
        foreach (var v in data.AllowedInputVariables) {
          var val = data.Dataset.GetDoubleValue(v, r);
          for (var j = 0; j < dimensions; j++)
            transformedTrainingset[rowCount, j] += val * transformationMatrix[i, j];
          i++;
        }
        rowCount++;
      }

      var ds = data.Dataset;
      var targetVariable = data.TargetVariable;
      return new NCAModel(transformedTrainingset, transformationMatrix, k, data.TargetVariable, data.AllowedInputVariables,
        data.TrainingIndices.Select(i => ds.GetDoubleValue(targetVariable, i)).ToArray());
    }

    private static void Gradient(double[] A, ref double func, double[] grad, object obj) {
      var info = (OptimizationInfo)obj;
      var data = info.Data;
      var classes = info.TargetValues;
      var instances = info.Instances;
      var attributes = info.Attributes;

      var AMatrix = new Matrix(A, A.Length / info.ReduceDimensions, info.ReduceDimensions);

      alglib.sparsematrix probabilities;
      alglib.sparsecreate(instances, instances, out probabilities);
      var transformedDistances = new double[instances];
      for (int i = 0; i < instances - 1; i++) {
        var iVector = new Matrix(GetRow(data, i), data.GetLength(1));
        var denom = 0.0;
        for (int k = 0; k < instances; k++) {
          if (k == i) continue;
          var kVector = new Matrix(GetRow(data, k));
          transformedDistances[k] = Math.Exp(-iVector.Multiply(AMatrix).Subtract(kVector.Multiply(AMatrix)).SquaredVectorLength());
          denom += transformedDistances[k];
        }
        if (denom > 1e-05) {
          for (int j = i + 1; j < instances; j++) {
            if (i == j) continue;
            var v = transformedDistances[j] / denom;
            alglib.sparseset(probabilities, i, j, v);
          }
        }
      }
      alglib.sparseconverttocrs(probabilities); // needed to enumerate in order (top-down and left-right)

      int t0 = 0, t1 = 0, r, c;
      double val;
      var pi = new double[instances];
      while (alglib.sparseenumerate(probabilities, ref t0, ref t1, out r, out c, out val)) {
        if (classes[r].IsAlmost(classes[c])) {
          pi[r] += val;
        }
      }

      var innerSum = new double[attributes, attributes];
      while (alglib.sparseenumerate(probabilities, ref t0, ref t1, out r, out c, out val)) {
        var vector = new Matrix(GetRow(data, r)).Subtract(new Matrix(GetRow(data, c)));
        vector.OuterProduct(vector).Multiply(2.0 * val * pi[r]).AddTo(innerSum);

        if (classes[r].IsAlmost(classes[c])) {
          vector.OuterProduct(vector).Multiply(-2.0 * val).AddTo(innerSum);
        }
      }

      func = -2.0 * pi.Sum();

      r = 0;
      var newGrad = AMatrix.Multiply(-2.0).Transpose().Multiply(new Matrix(innerSum)).Transpose();
      foreach (var g in newGrad) {
        grad[r++] = g;
      }
    }

    #region Helpers
    private static IEnumerable<double> GetRow(double[,] data, int row) {
      for (int i = 0; i < data.GetLength(1); i++)
        yield return data[row, i];
    }

    private class OptimizationInfo {
      //public IClassificationProblemData ProblemData { get; private set; }
      public Scaling Scaling { get; private set; }
      public double[,] Data { get; private set; }
      public double[] TargetValues { get; private set; }
      public int ReduceDimensions { get; private set; }
      public int Instances { get; private set; }
      public int Attributes { get; private set; }

      public OptimizationInfo(IClassificationProblemData data, int reduceDimensions) {
        //this.ProblemData = data;
        this.Scaling = new Scaling(data.Dataset, data.AllowedInputVariables, data.TrainingIndices);
        this.Data = AlglibUtil.PrepareAndScaleInputMatrix(data.Dataset, data.AllowedInputVariables, data.TrainingIndices, Scaling);
        this.TargetValues = data.Dataset.GetDoubleValues(data.TargetVariable, data.TrainingIndices).ToArray();
        this.ReduceDimensions = reduceDimensions;
        this.Instances = data.TrainingIndices.Count();
        this.Attributes = data.AllowedInputVariables.Count();
      }
    }
    #endregion
  }
}
